const Ast = @import("./kym/Ast.zig");

const Chunk = @import("./kym/Chunk.zig");

const Table = @import("./kym/Table.zig");

const coral = @import("coral");

const file = @import("./file.zig");

const tokens = @import("./kym/tokens.zig");

pub const CallContext = struct {
	env: *RuntimeEnv,
	caller: ?*const RuntimeRef,
	userdata: []u8,
	args: []const ?*const RuntimeRef = &.{},

	pub fn arg_at(self: CallContext, index: u8) RuntimeError!*const RuntimeRef {
		if (!coral.math.is_clamped(index, 0, self.args.len - 1)) {
			return self.env.check_fail("argument out of bounds");
		}

		return self.args[@as(usize, index)];
	}
};

pub const DynamicObject = struct {
	userdata: []coral.io.Byte,
	typeinfo: *const Typeinfo,
};

pub const Float = f64;

pub const IndexContext = struct {
	env: *RuntimeEnv,
	userdata: []coral.io.Byte,
	index_ref: ?*const RuntimeRef,
};

pub const RuntimeEnv = struct {
	allocator: coral.io.Allocator,
	err_writer: coral.io.Writer,
	local_refs: RefStack,
	frames: FrameStack,
	ref_values: RefSlab,

	const FrameStack = coral.list.Stack(struct {
		locals_top: usize,
	});

	const RefStack = coral.list.Stack(?*RuntimeRef);

	pub const Options = struct {
		out_writer: coral.io.Writer = coral.io.null_writer,
		err_writer: coral.io.Writer = coral.io.null_writer,
	};

	const RefSlab = coral.map.Slab(struct {
		ref_count: usize,

		object: union (enum) {
			false,
			true,
			number: Float,
			string: []coral.io.Byte,
			dynamic: *DynamicObject,
		},
	});

	pub const ScriptSource = struct {
		name: []const coral.io.Byte,
		data: []const coral.io.Byte,
	};

	pub fn acquire(self: *RuntimeEnv, ref: ?*const RuntimeRef) ?*RuntimeRef {
		const key = @intFromPtr(ref);
		var ref_data = self.ref_values.remove(key);

		coral.debug.assert(ref_data != null);

		ref_data.?.ref_count += 1;

		coral.debug.assert(self.ref_values.insert_at(key, ref_data.?));

		return @ptrFromInt(key);
	}

	pub fn call(
		self: *RuntimeEnv,
		caller_ref: *const RuntimeRef,
		callable_ref: *const RuntimeRef,
		arg_refs: []const *const RuntimeRef,
	) RuntimeError!*RuntimeRef {
		const callable = (try callable_ref.fetch(self)).to_object();

		return callable.userinfo.call(.{
			.env = self,
			.caller = caller_ref,
			.callable = callable_ref,
			.userdata = callable.userdata,
			.args = arg_refs,
		});
	}

	pub fn discard(self: *RuntimeEnv, ref: *?*RuntimeRef) void {
		const key = @intFromPtr(ref.* orelse return);
		var ref_data = self.ref_values.remove(key) orelse unreachable;

		coral.debug.assert(ref_data.ref_count != 0);

		ref_data.ref_count -= 1;

		if (ref_data.ref_count == 0) {
			switch (ref_data.object) {
				.string => |string| self.allocator.deallocate(string),
				.dynamic => |virtual| self.allocator.deallocate(virtual),
				else => {},
			}
		} else {
			coral.debug.assert(self.ref_values.insert_at(key, ref_data));
		}

		ref.* = null;
	}

	pub fn execute_file(self: *RuntimeEnv, global_ref: ?*const RuntimeRef, file_access: file.Access, file_path: file.Path) RuntimeError!?*RuntimeRef {
		const error_message = "failed to load file";

		const file_data = (try file.allocate_and_load(self.allocator, file_access, file_path)) orelse {
			return self.raise(error.SystemFailure, error_message);
		};

		defer self.allocator.deallocate(file_data);

		return self.execute_script(global_ref, .{
			.name = file_path.to_string() orelse return self.raise(error.SystemFailure, error_message),
			.data = file_data,
		});
	}

	pub fn execute_script(self: *RuntimeEnv, global_ref: ?*const RuntimeRef, source: ScriptSource) RuntimeError!?*RuntimeRef {
		var ast = Ast.make(self.allocator);

		defer ast.free();

		{
			var tokenizer = tokens.Tokenizer{.source = source.data};

			ast.parse(&tokenizer) catch |parse_error| switch (parse_error) {
				error.BadSyntax => return self.raise(error.BadSyntax, ast.error_message),
				error.OutOfMemory => return error.OutOfMemory,
			};
		}

		var exe = Chunk.make(self);

		defer exe.free();

		try exe.compile_ast(ast);

		return exe.execute(global_ref, source.name);
	}

	pub fn frame_pop(self: *RuntimeEnv) RuntimeError!void {
		const frame = self.frames.pop() orelse return self.raise(error.IllegalState, "stack underflow");

		coral.debug.assert(self.local_refs.drop(self.local_refs.values.len - frame.locals_top));
	}

	pub fn frame_push(self: *RuntimeEnv) RuntimeError!void {
		return self.frames.push_one(.{
			.locals_top = self.local_refs.values.len,
		});
	}

	pub fn free(self: *RuntimeEnv) void {
		self.local_refs.free();
		self.ref_values.free();
	}

	pub fn get_dynamic(self: *RuntimeEnv, indexable_ref: ?*const RuntimeRef, index_ref: ?*const RuntimeRef) RuntimeError!?*RuntimeRef {
		return switch (self.unbox(indexable_ref)) {
			.nil => self.raise(error.BadOperation, "nil is immutable"),
			.boolean => self.raise(error.BadOperation, "boolean has no such index"),
			.number => self.raise(error.BadOperation, "number has no such index"),
			.string => self.raise(error.BadOperation, "string has no such index"),

			.dynamic => |dynamic| dynamic.typeinfo.get(.{
				.userdata = dynamic.userdata,
				.env = self,
				.index_ref = index_ref,
			}),
		};
	}

	pub fn local_get(self: *RuntimeEnv, local: u8) RuntimeError!?*RuntimeRef {
		return self.local_refs.values[local];
	}

	pub fn local_pop(self: *RuntimeEnv) RuntimeError!?*RuntimeRef {
		return self.local_refs.pop() orelse self.raise(error.IllegalState, "stack underflow");
	}

	pub fn local_push_ref(self: *RuntimeEnv, ref: ?*RuntimeRef) RuntimeError!void {
		return self.local_refs.push_one(self.acquire(ref));
	}

	pub fn local_push_boolean(self: *RuntimeEnv, boolean: bool) RuntimeError!void {
		return self.local_refs.push_one(try self.new_boolean(boolean));
	}

	pub fn local_push_number(self: *RuntimeEnv, number: Float) RuntimeError!void {
		return self.local_refs.push_one(try self.new_number(number));
	}

	pub fn local_set(self: *RuntimeEnv, local: u8, value: ?*RuntimeRef) RuntimeError!void {
		self.local_refs.values[local] = self.acquire(value);
	}

	pub fn make(allocator: coral.io.Allocator, options: Options) RuntimeError!RuntimeEnv {
		return .{
			.local_refs = RefStack.make(allocator),
			.ref_values = RefSlab.make(allocator),
			.frames = FrameStack.make(allocator),
			.err_writer = options.err_writer,
			.allocator = allocator,
		};
	}

	pub fn new_boolean(self: *RuntimeEnv, value: bool) RuntimeError!?*RuntimeRef {
		return @ptrFromInt(try self.ref_values.insert(.{
			.ref_count = 1,
			.object = if (value) .true else .false,
		}));
	}

	pub fn new_dynamic(self: *RuntimeEnv, userdata: []const coral.io.Byte, typeinfo: *const Typeinfo) RuntimeError!*RuntimeRef {
		const userdata_copy = try coral.io.allocate_copy(self.allocator, userdata);

		errdefer self.allocator.deallocate(userdata_copy);

		const dynamic_object = try coral.io.allocate_one(self.allocator, DynamicObject{
			.typeinfo = typeinfo,
			.userdata = userdata_copy,
		});

		errdefer self.allocator.deallocate(dynamic_object);

		return @ptrFromInt(try self.ref_values.insert(.{
			.ref_count = 1,
			.object = .{.dynamic = dynamic_object},
		}));
	}

	pub fn new_number(self: *RuntimeEnv, value: Float) RuntimeError!?*RuntimeRef {
		return @ptrFromInt(try self.ref_values.insert(.{
			.ref_count = 1,
			.object = .{.number = value},
		}));
	}

	pub fn new_string(self: *RuntimeEnv, string_data: []const coral.io.Byte) RuntimeError!?*RuntimeRef {
		const string_copy = try coral.io.allocate_copy(self.allocator, string_data);

		errdefer self.allocator.deallocate(string_copy);

		return @ptrFromInt(try self.ref_values.insert(.{
			.ref_count = 1,
			.object = .{.string = string_copy},
		}));
	}

	pub fn raise(self: *RuntimeEnv, error_value: RuntimeError, error_message: []const u8) RuntimeError {
		// TODO: Print stack trace from state.
		coral.utf8.print_formatted(self.err_writer, "{name}@{line}: {message}", .{
			.name = "???",
			.line = @as(u64, 0),
			.message = error_message,
		}) catch {};

		return error_value;
	}

	pub fn set_dynamic(self: *RuntimeEnv, indexable_ref: ?*const RuntimeRef, index_ref: ?*const RuntimeRef, value_ref: ?*const RuntimeRef) RuntimeError!void {
		return switch (self.unbox(indexable_ref)) {
			.nil => self.raise(error.BadOperation, "nil is immutable"),
			.boolean => self.raise(error.BadOperation, "boolean is immutable"),
			.number => self.raise(error.BadOperation, "number is immutable"),
			.string => self.raise(error.BadOperation, "string is immutable"),

			.dynamic => |dynamic| dynamic.typeinfo.set(.{
				.userdata = dynamic.userdata,
				.env = self,
				.index_ref = index_ref,
			}, value_ref),
		};
	}

	pub fn unbox(self: *RuntimeEnv, ref: ?*const RuntimeRef) Unboxed {
		const ref_data = self.ref_values.lookup(@intFromPtr(ref orelse return .nil));

		coral.debug.assert(ref_data != null);

		return switch (ref_data.?.object) {
			.false => .{.boolean = false},
			.true => .{.boolean = true},
			.number => |number| .{.number = number},
			.string => |string| .{.string = string},
			.dynamic => |dynamic| .{.dynamic = dynamic},
		};
	}
};

pub const RuntimeError = coral.io.AllocationError || error {
	IllegalState,
	SystemFailure,
	TypeMismatch,
	BadOperation,
	BadSyntax,
};

pub const RuntimeRef = opaque {};

pub const TestContext = struct {
	env: *RuntimeEnv,
	userdata: []const coral.io.Byte,
	other_ref: ?*const RuntimeRef,
};

pub const Typeinfo = struct {
	call: *const fn (context: CallContext) RuntimeError!?*RuntimeRef,
	clean: *const fn (userdata: []u8) void,
	get: *const fn (context: IndexContext) RuntimeError!?*RuntimeRef,
	set: *const fn (context: IndexContext, value: ?*const RuntimeRef) RuntimeError!void,
	test_difference: *const fn (context: TestContext) RuntimeError!Float = default_test_difference,
	test_equality: *const fn (context: TestContext) RuntimeError!bool = default_test_equality,

	fn default_test_difference(context: TestContext) RuntimeError!Float {
		return context.env.raise(error.TypeMismatch, "object is not comparable");
	}

	fn default_test_equality(context: TestContext) RuntimeError!bool {
		return switch (context.env.unbox(context.other_ref)) {
			.dynamic => |dynamic| context.userdata.ptr == dynamic.userdata.ptr,
			else => false,
		};
	}
};

pub const Unboxed = union (enum) {
	nil,
	boolean: bool,
	number: Float,
	string: []const u8,
	dynamic: *const DynamicObject,
};

pub fn get_dynamic_field(env: *RuntimeEnv, indexable: ?*const RuntimeRef, field: []const u8) RuntimeError!?*RuntimeRef {
	var interned_field = try env.new_string(field);

	defer env.discard(&interned_field);

	return env.get_dynamic(indexable, interned_field);
}

pub fn new_table(env: *RuntimeEnv) RuntimeError!?*RuntimeRef {
	var table = Table.make(env);

	errdefer table.free();

	return try env.new_dynamic(coral.io.bytes_of(&table), &Table.typeinfo);
}

pub fn test_difference(env: *RuntimeEnv, lhs_ref: ?*const RuntimeRef, rhs_ref: ?*const RuntimeRef) RuntimeError!Float {
	return switch (env.unbox(lhs_ref)) {
		.nil => env.raise(error.TypeMismatch, "cannot compare nil objects"),
		.boolean => env.raise(error.TypeMismatch, "cannot compare boolean objects"),

		.number => |lhs_number| switch (env.unbox(rhs_ref)) {
			.number => |rhs_number| rhs_number - lhs_number,
			else => env.raise(error.TypeMismatch, "right-hand object is not comparable with number objects"),
		},

		.string => |lhs_string| switch (env.unbox(rhs_ref)) {
			.string => |rhs_string| @floatFromInt(coral.io.compare(lhs_string, rhs_string)),
			else => env.raise(error.TypeMismatch, "right-hand object is not comparable with string objects"),
		},

		.dynamic => |lhs_dynamic| lhs_dynamic.typeinfo.test_difference(.{
			.env = env,
			.userdata = lhs_dynamic.userdata,
			.other_ref = rhs_ref,
		}),
	};
}

pub fn test_equality(env: *RuntimeEnv, lhs_ref: ?*const RuntimeRef, rhs_ref: ?*const RuntimeRef) RuntimeError!bool {
	return switch (env.unbox(lhs_ref)) {
		.nil => (rhs_ref == null),

		.boolean => |lhs_boolean| switch (env.unbox(rhs_ref)) {
			.boolean => |rhs_boolean| rhs_boolean == lhs_boolean,
			else => false,
		},

		.number => |lhs_number| switch (env.unbox(rhs_ref)) {
			.number => |rhs_number| lhs_number == rhs_number,
			else => false,
		},

		.string => |lhs_string| switch (env.unbox(rhs_ref)) {
			.string => |rhs_string| coral.io.equals(lhs_string, rhs_string),
			else => false,
		},

		.dynamic => |lhs_dynamic| try lhs_dynamic.typeinfo.test_equality(.{
			.env = env,
			.userdata = lhs_dynamic.userdata,
			.other_ref = rhs_ref,
		}),
	};
}
